/*
 * Copyright 2017 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.api.internal.artifacts.ivyservice.resolveengine.artifact;

import org.gradle.api.artifacts.ResolvedArtifact;
import org.gradle.api.artifacts.component.ComponentArtifactIdentifier;
import org.gradle.api.internal.tasks.TaskDependencyContainer;
import org.gradle.internal.model.CalculatedValue;

import java.io.File;

/**
 * Represents an artifact that can be resolved. Call {@link #getFile()} or {@link ResolvedArtifact#getFile()} to resolve.
 */
public interface ResolvableArtifact extends TaskDependencyContainer {
    ComponentArtifactIdentifier getId();

    /**
     * Should this artifact be resolved synchronously? For example, is the result of {@link #getFile()} available in memory or can it be calculated quickly without IO calls?
     */
    boolean isResolveSynchronously();

    /**
     * Resolves the file, if not already, blocking until complete.
     */
    File getFile();

    /**
     * Returns the artifact file as a lazy type. Does not resolve the file, but the returned value can be used to do so.
     */
    CalculatedValue<File> getFileSource();

    ResolvableArtifact transformedTo(File file);

    ResolvedArtifact toPublicView();
}
