/*
 * Copyright 2022 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.api.tasks.diagnostics.internal.dependencies;

public enum MatchType {
    /**
     * The attribute was not requested by the configuration.
     */
    NOT_REQUESTED,
    /**
     * The attribute was requested by the configuration, but the actual value was <strong>incompatible</strong> with
     * the configuration's requested value.
     */
    INCOMPATIBLE,
    /**
     * The attribute was requested by the configuration, and the actual value was <strong>compatible</strong> with
     * the configuration's requested value.
     */
    COMPATIBLE,
    /**
     * The attribute was requested by the configuration, and the actual value was <strong>equal</strong> to the
     * configuration's requested value.
     */
    EQUAL,
}
