/*
 * Copyright 2015 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.api.internal.artifacts.ivyservice.resolveengine.projectresult;

import org.gradle.api.artifacts.component.ProjectComponentIdentifier;

public class DefaultResolvedProjectConfiguration implements ResolvedProjectConfiguration {
    private final ProjectComponentIdentifier id;
    private final String targetConfiguration;

    public DefaultResolvedProjectConfiguration(ProjectComponentIdentifier id, String targetConfiguration) {
        this.id = id;
        this.targetConfiguration = targetConfiguration;
    }

    @Override
    public ProjectComponentIdentifier getId() {
        return id;
    }

    @Override
    public String getTargetConfiguration() {
        return targetConfiguration;
    }

    @Override
    public String toString() {
        return id + ":" + targetConfiguration;
    }
}
