/*
 * Copyright 2022 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.internal.scripts;

public class ScriptOriginUtil {

    /**
     * Extracts the class identifier of the value defined in a build script
     * independent of its absolute path.
     *
     * The identifier depends on the original class name and the content hash of the script.
     * As such it can be used for build caching between different locations.
     */
    public static String getOriginClassIdentifier(Object value) {
        if (value instanceof ScriptOrigin) {
            ScriptOrigin origin = (ScriptOrigin) value;
            return origin.getOriginalClassName() + "_" + origin.getContentHash();
        } else {
            return value.getClass().getName();
        }
    }
}
