/*
 * Copyright 2019 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.api.internal.tasks;

import javax.annotation.Nullable;

public class DefaultTaskDependencyFactory implements TaskDependencyFactory {
    @Nullable
    private final TaskResolver taskResolver;

    public static TaskDependencyFactory withNoAssociatedProject() {
        return new DefaultTaskDependencyFactory(null);
    }

    public static TaskDependencyFactory forProject(TaskResolver taskResolver) {
        return new DefaultTaskDependencyFactory(taskResolver);
    }

    private DefaultTaskDependencyFactory(@Nullable TaskResolver taskResolver) {
        this.taskResolver = taskResolver;
    }

    @Override
    public DefaultTaskDependency configurableDependency() {
        return new DefaultTaskDependency(taskResolver);
    }
}
