/*
 * Copyright 2014 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.cache;

import org.gradle.internal.Factory;

public interface CrossProcessCacheAccess {
    /**
     * Runs the given action while this process is holding an exclusive file lock on the cache. Multiple threads may run concurrently.
     */
    <T> T withFileLock(Factory<T> factory);

    /**
     * Acquires an exclusive file lock on the cache. The caller is responsible for running the resulting action to release the lock.
     * The lock may be released by any thread.
     */
    Runnable acquireFileLock();
}
