/*
 * Copyright 2019 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.gradle.api.internal.artifacts.ivyservice.ivyresolve.verification;

import com.google.common.collect.ImmutableList;
import org.gradle.internal.UncheckedException;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;

public abstract class DefaultKeyServers {
    private final static List<URI> DEFAULT_KEYSERVERS = ImmutableList.of(
        uri("hkp://ha.pool.sks-keyservers.net"),
        uri("https://keyserver.ubuntu.com"),
        uri("https://keys.openpgp.org"),
        uri("https://pgp.mit.edu")
    );

    private static URI uri(String uri) {
        try {
            return new URI(uri);
        } catch (URISyntaxException e) {
            throw UncheckedException.throwAsUncheckedException(e);
        }
    }

    public static List<URI> getOrDefaults(List<URI> uris) {
        if (uris.isEmpty()) {
            return DEFAULT_KEYSERVERS;
        }
        return uris;
    }
}
