/*
 * Copyright 2021 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.api.internal.file.collections;

import org.gradle.api.file.FileCollection;

/**
 * A {@link FileCollection} implementation that throws a given exception when visited.
 */
public class FailingFileCollection extends LazilyInitializedFileCollection {

    private final String displayName;
    private final RuntimeException failure;

    public FailingFileCollection(String displayName, RuntimeException failure) {
        this.displayName = displayName;
        this.failure = failure;
    }

    @Override
    public String getDisplayName() {
        return displayName;
    }

    @Override
    public FileCollection createDelegate() {
        throw failure;
    }
}
