/*
 * Copyright 2021 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.composite.internal;

import org.gradle.api.artifacts.component.BuildIdentifier;
import org.gradle.api.internal.TaskInternal;

public interface TaskIdentifier {
    BuildIdentifier getBuildIdentifier();

    String getTaskPath();

    interface TaskBasedTaskIdentifier extends TaskIdentifier {
        TaskInternal getTask();
    }

    static TaskBasedTaskIdentifier of(BuildIdentifier buildIdentifier, TaskInternal task) {
        return new TaskBasedTaskIdentifier() {
            @Override
            public BuildIdentifier getBuildIdentifier() {
                return buildIdentifier;
            }

            @Override
            public TaskInternal getTask() {
                return task;
            }

            @Override
            public String getTaskPath() {
                return task.getPath();
            }
        };
    }

    static TaskIdentifier of(BuildIdentifier buildIdentifier, String taskPath) {
        return new TaskIdentifier() {
            @Override
            public BuildIdentifier getBuildIdentifier() {
                return buildIdentifier;
            }

            @Override
            public String getTaskPath() {
                return taskPath;
            }
        };
    }
}
