/*
 * Copyright 2018 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.tooling.events;

import java.net.URI;

/**
 * Identifies a Gradle script plugin.
 *
 * @since 5.1
 */
public interface ScriptPluginIdentifier extends PluginIdentifier {

    /**
     * Returns the URI of this script plugin.
     *
     * <p>The URI may point to the local file system or a remote server.
     */
    URI getUri();

}
