/*
 * Copyright 2018 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.internal.execution.history;

import com.google.common.collect.ImmutableSortedMap;
import org.gradle.caching.internal.origin.OriginMetadata;
import org.gradle.internal.fingerprint.FileCollectionFingerprint;

/**
 * Captures the state a {@link org.gradle.internal.execution.UnitOfWork} after the previous execution has finished.
 */
public interface PreviousExecutionState extends InputExecutionState, OutputExecutionState {

    /**
     * The ID and execution time of origin of the execution's outputs.
     */
    OriginMetadata getOriginMetadata();

    /**
     * Whether the execution was successful.
     */
    boolean isSuccessful();

    /**
     * {@inheritDoc}
     */
    @Override
    ImmutableSortedMap<String, FileCollectionFingerprint> getInputFileProperties();
}
