/*
 * Copyright 2016 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.gradle.internal.normalization.java.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class ClassMember extends AnnotatableMember {

    private final int version;
    private final String superName;
    private final String[] interfaces;

    private final List<String> permittedSubclasses = new ArrayList<>();

    public ClassMember(int version, int access, String name, String signature, String superName, String[] interfaces) {
        super(access, name, signature);
        this.version = version;
        this.superName = superName;
        this.interfaces = interfaces;
    }

    public String[] getInterfaces() {
        return Arrays.copyOf(interfaces, interfaces.length);
    }

    public String getSuperName() {
        return superName;
    }

    public int getVersion() {
        return version;
    }

    public List<String> getPermittedSubclasses() {
        return permittedSubclasses;
    }
}
